@@ -4,14 +4,27 @@ from __future__ import annotations
 import argparse
 import json
 import os
+import shlex
+import shutil
 import sys
+import tempfile
+import re
+
+try:  # pragma: no cover - fallback only for older Python
+    from importlib import metadata as importlib_metadata
+except ImportError:  # pragma: no cover
+    try:
+        import importlib_metadata  # type: ignore
+    except ImportError:
+        importlib_metadata = None  # type: ignore
+
 from importlib import resources
 from pathlib import Path
 from typing import IO, Dict, List, Optional, Sequence, Tuple
 
 import requests
 import yaml
-from dotenv import load_dotenv
+from dotenv import dotenv_values, find_dotenv
 from openai import OpenAI, OpenAIError
 
 YELLOW = "\033[93m"
@@ -22,8 +35,55 @@ _DEFAULT_SYSTEM_PROMPT_RESOURCE = "data/bsy-clippy.txt"
 _DEFAULT_CONFIG_FILENAME = "bsy-clippy.yaml"
 _DEFAULT_OPENAI_BASE_URL = "https://api.openai.com/v1"
 
-load_dotenv()
-
+MAX_CONTEXT_TOKENS = 65_536
+AVERAGE_CHARS_PER_TOKEN = 4
+MAX_CONTEXT_CHARS = MAX_CONTEXT_TOKENS * AVERAGE_CHARS_PER_TOKEN
+TOOL_OUTPUT_SUMMARY_CHAR_LIMIT = 16_384
+USER_INPUT_SUMMARY_CHAR_LIMIT = 60_000
+DEFAULT_SUMMARY_CHUNK_SIZE = 4_096
+MIN_SUMMARY_CHUNK_SIZE = 512
+SUMMARY_SEPARATOR = "\n---\n"
+SUMMARY_NOTICE = "[Tool output summarized]"
+USER_SUMMARY_NOTICE = "[User input summarized]"
+MAX_SUMMARY_ITERATIONS = 6
+HELPER_COMMAND_PREFIXES = {"sudo", "doas", "env"}
+DANGEROUS_PRIMARY_COMMANDS = {
+    "rm",
+    "rmdir",
+    "unlink",
+    "del",
+    "erase",
+    "dd",
+    "shred",
+    "srm",
+    "wipefs",
+    "mkfs",
+    "mkfs.ext4",
+    "mkfs.ext3",
+    "mkfs.ntfs",
+    "mkfs.fat",
+    "mkfs.vfat",
+    "mkdosfs",
+    "diskutil",
+    "format",
+    "rd",
+}
+DANGEROUS_FLAGS = {
+    "-rf",
+    "-fr",
+    "-delete",
+    "--delete",
+    "--no-preserve-root",
+}
+DANGEROUS_PHRASES = (
+    "rm -rf",
+    "rm -fr",
+    "rm -r ",
+    "rm -f ",
+    "git clean",
+    "git reset --hard",
+    "xargs rm",
+)
 
 # ---------------------------------------------------------------------------
 # Utility helpers
@@ -69,6 +129,59 @@ def compose_prompt(*parts: Optional[str]) -> str:
     return "\n\n".join(collected)
 
 
+def get_app_version() -> str:
+    if importlib_metadata is not None:
+        try:
+            return importlib_metadata.version("bsy-clippy")
+        except Exception:
+            pass
+
+    project_path = Path(__file__).resolve().parent.parent / "pyproject.toml"
+    try:
+        text = project_path.read_text(encoding="utf-8")
+    except OSError:
+        text = ""
+
+    if text:
+        toml_loader = None
+        try:
+            import tomllib  # type: ignore[attr-defined]
+
+            toml_loader = tomllib  # type: ignore[assignment]
+        except ModuleNotFoundError:
+            toml_loader = None
+
+        if toml_loader:
+            try:
+                data = toml_loader.loads(text)
+                project_data = data.get("project")
+                if isinstance(project_data, dict):
+                    version = project_data.get("version")
+                    if version:
+                        return str(version)
+            except Exception:
+                pass
+
+        match = re.search(r'^\s*version\s*=\s*"([^"]+)"\s*$', text, re.MULTILINE)
+        if match:
+            return match.group(1)
+
+    return "unknown"
+
+
+def build_version_banner(version: str, extra: Optional[str] = None) -> str:
+    banner = f"bsy-clippy {version}"
+    if extra:
+        banner = f"{banner} | {extra}"
+    return banner
+
+
+def print_version_banner(version: Optional[str] = None, extra: Optional[str] = None) -> None:
+    if version is None:
+        version = get_app_version()
+    print(build_version_banner(version, extra))
+
+
 def strip_think_segments(text: str) -> str:
     if not text:
         return ""
@@ -124,6 +237,153 @@ def colorize_response(text: str) -> str:
     return "".join(output)
 
 
+def _split_text_into_chunks(text: str, chunk_size: int) -> List[str]:
+    chunk_size = max(chunk_size, MIN_SUMMARY_CHUNK_SIZE)
+    return [text[i : i + chunk_size] for i in range(0, len(text), chunk_size)]
+
+
+def _summarize_chunk_text(chunk: str) -> str:
+    snippet = chunk.strip()
+    if not snippet:
+        return ""
+
+    if len(snippet) <= MIN_SUMMARY_CHUNK_SIZE:
+        return snippet
+
+    lines = snippet.splitlines()
+    if len(lines) <= 8:
+        return snippet
+
+    head = lines[:3]
+    tail = lines[-3:]
+    omitted = max(0, len(lines) - len(head) - len(tail))
+
+    summary: List[str] = []
+    summary.extend(head)
+    if omitted > 0:
+        summary.append(f"... ({omitted} more lines omitted) ...")
+    summary.extend(tail)
+
+    condensed = "\n".join(summary).strip()
+    if not condensed:
+        return snippet[:MIN_SUMMARY_CHUNK_SIZE]
+    return condensed
+
+
+def _summarize_iteratively(text: str, limit: int, chunk_size: int = DEFAULT_SUMMARY_CHUNK_SIZE) -> Tuple[str, bool]:
+    current = text.strip()
+    if len(current) <= limit:
+        return current, False
+
+    chunk_size = max(chunk_size, MIN_SUMMARY_CHUNK_SIZE)
+    iteration = 0
+
+    while len(current) > limit and chunk_size >= MIN_SUMMARY_CHUNK_SIZE and iteration < MAX_SUMMARY_ITERATIONS:
+        chunks = _split_text_into_chunks(current, chunk_size)
+        chunk_summaries: List[str] = []
+        total = len(chunks)
+        for index, chunk in enumerate(chunks, start=1):
+            summary = _summarize_chunk_text(chunk)
+            if not summary:
+                continue
+            header = f"[Chunk {index}/{total}]"
+            chunk_summaries.append(f"{header}\n{summary}")
+        condensed = SUMMARY_SEPARATOR.join(chunk_summaries).strip()
+        if not condensed or len(condensed) >= len(current):
+            break
+        current = condensed
+        chunk_size = max(chunk_size // 2, MIN_SUMMARY_CHUNK_SIZE)
+        iteration += 1
+
+    did_truncate = len(current) < len(text.strip())
+
+    if len(current) > limit:
+        trimmed = current[:limit]
+        if "\n" in trimmed:
+            trimmed = trimmed.rsplit("\n", 1)[0]
+        current = f"{trimmed.rstrip()}\n[...truncated due to context limit...]"
+        did_truncate = True
+
+    return current.strip(), did_truncate
+
+
+def summarize_tool_output(text: str) -> Tuple[str, bool, Tuple[int, int]]:
+    snippet = text.strip()
+    if not snippet:
+        return "", False, (0, 0)
+    summary, truncated = _summarize_iteratively(snippet, TOOL_OUTPUT_SUMMARY_CHAR_LIMIT)
+    line_count = snippet.count("\n") + 1
+    return summary, truncated, (line_count, len(snippet))
+
+
+def summarize_user_input(text: str) -> Tuple[str, bool, Tuple[int, int]]:
+    snippet = text.strip()
+    if not snippet:
+        return "", False, (0, 0)
+    summary, truncated = _summarize_iteratively(snippet, USER_INPUT_SUMMARY_CHAR_LIMIT)
+    line_count = snippet.count("\n") + 1
+    return summary, truncated, (line_count, len(snippet))
+
+
+def _is_command_safe(cmd: str, parms: Sequence[str]) -> Tuple[bool, Optional[str]]:
+    tokens = [str(cmd)] + [str(item) for item in parms]
+    cleaned: List[Tuple[str, str, str]] = []
+    for token in tokens:
+        token_str = token.strip()
+        if not token_str:
+            continue
+        lowered = token_str.lower()
+        base = os.path.basename(lowered)
+        cleaned.append((token_str, lowered, base))
+
+    if not cleaned:
+        return False, "no command provided"
+
+    primary_index = 0
+    for idx, (_, lowered, base) in enumerate(cleaned):
+        if base in HELPER_COMMAND_PREFIXES:
+            continue
+        primary_index = idx
+        break
+
+    primary_lower = cleaned[primary_index][1]
+    primary_base = cleaned[primary_index][2]
+
+    if primary_base in DANGEROUS_PRIMARY_COMMANDS:
+        return False, f"command '{primary_base}' is blocked"
+
+    for idx, (_, lowered, base) in enumerate(cleaned):
+        if idx != primary_index and base in DANGEROUS_PRIMARY_COMMANDS:
+            return False, f"sub-command '{base}' is blocked"
+        if lowered in DANGEROUS_FLAGS:
+            return False, f"flag '{lowered}' is blocked"
+
+    if primary_base == "git":
+        subcommand = ""
+        if primary_index + 1 < len(cleaned):
+            subcommand = cleaned[primary_index + 1][2]
+        if subcommand in {"rm", "clean"}:
+            return False, f"'git {subcommand}' operations are blocked"
+        if subcommand == "reset":
+            forbidden = {"--hard", "--merge", "--keep", "--mixed", "--soft"}
+            for _, lowered, _ in cleaned[primary_index + 2 :]:
+                if lowered in forbidden:
+                    return False, "dangerous git reset option is blocked"
+
+    for idx, (_, lowered, base) in enumerate(cleaned):
+        if idx == primary_index:
+            continue
+        if base in {"rm", "rmdir", "unlink", "del"}:
+            return False, f"sub-command '{base}' is blocked"
+
+    combined = " ".join(lowered for _, lowered, _ in cleaned)
+    for phrase in DANGEROUS_PHRASES:
+        if phrase in combined:
+            return False, f"pattern '{phrase}' is blocked"
+
+    return True, None
+
+
 def print_stream_chunk(text: str, in_think: bool) -> Tuple[bool, str]:
     idx = 0
     final_parts: List[str] = []
@@ -472,6 +732,205 @@ def read_user_input(prompt_text: str, input_stream: Optional[IO[str]]) -> str:
     return line.rstrip("\r\n")
 
 
+def _extract_tool_requests(raw_text: str) -> Tuple[str, List[Dict[str, object]]]:
+    if not raw_text:
+        return "", []
+    decoder = json.JSONDecoder()
+    idx = 0
+    last_idx = 0
+    length = len(raw_text)
+    requests: List[Dict[str, object]] = []
+    sanitized_segments: List[str] = []
+
+    while idx < length:
+        brace_idx = raw_text.find("{", idx)
+        if brace_idx == -1:
+            break
+        try:
+            obj, end_idx = decoder.raw_decode(raw_text, brace_idx)
+        except ValueError:
+            idx = brace_idx + 1
+            continue
+
+        sanitized_segments.append(raw_text[last_idx:brace_idx])
+        idx = end_idx
+
+        if not isinstance(obj, dict):
+            sanitized_segments.append(raw_text[brace_idx:end_idx])
+            last_idx = end_idx
+            continue
+
+        cmd_value = obj.get("cmd")
+        if cmd_value is None:
+            sanitized_segments.append(raw_text[brace_idx:end_idx])
+            last_idx = end_idx
+            continue
+
+        cmd_str = str(cmd_value).strip()
+        if not cmd_str:
+            sanitized_segments.append(raw_text[brace_idx:end_idx])
+            last_idx = end_idx
+            continue
+
+        parms_value = obj.get("parms")
+        if parms_value is None and "params" in obj:
+            parms_value = obj.get("params")
+
+        if isinstance(parms_value, list):
+            parms_list = [str(item) for item in parms_value]
+        elif parms_value in (None,):
+            parms_list = []
+        else:
+            parms_list = [str(parms_value)]
+
+        requests.append({"cmd": cmd_str, "parms": parms_list})
+        last_idx = end_idx
+
+    sanitized_segments.append(raw_text[last_idx:])
+    sanitized_text = "".join(sanitized_segments).strip()
+    return sanitized_text, requests
+
+
+def _prompt_yes_no(prompt_text: str, input_stream: Optional[IO[str]]) -> bool:
+    current_prompt = prompt_text
+    while True:
+        try:
+            response = read_user_input(current_prompt, input_stream)
+        except EOFError:
+            print("\n[Tool] No confirmation received; skipping command.")
+            return False
+        except KeyboardInterrupt:
+            print("\n[Tool] Command cancelled by user.")
+            return False
+
+        normalized = response.strip().lower()
+        if normalized in {"y", "yes"}:
+            return True
+        if normalized in {"", "n", "no"}:
+            return False
+        current_prompt = "Please respond with 'y' or 'n': "
+
+
+def _interpret_exit_status(status: int) -> int:
+    if status < 0:
+        return status
+    signal = status & 0x7F
+    if signal:
+        return -signal
+    return (status >> 8) & 0xFF
+
+
+def _execute_tool_command(cmd: str, parms: Sequence[str]) -> Tuple[str, str, int]:
+    args = [str(cmd)] + [str(item) for item in parms]
+    command_display = shlex.join(args)
+    tee_available = shutil.which("tee") is not None and os.name != "nt"
+
+    temp_handle = tempfile.NamedTemporaryFile(delete=False)
+    temp_handle.close()
+    temp_path = temp_handle.name
+
+    try:
+        if tee_available:
+            shell_command = f"{command_display} 2>&1 | tee {shlex.quote(temp_path)}"
+        else:
+            shell_command = f"{command_display} > {shlex.quote(temp_path)} 2>&1"
+
+        print(f"[Tool] Executing: {command_display}")
+        status = os.system(shell_command)
+
+        try:
+            output_text = Path(temp_path).read_text(encoding="utf-8", errors="replace")
+        except OSError:
+            output_text = ""
+
+        if not tee_available and output_text:
+            print(output_text, end="" if output_text.endswith("\n") else "\n")
+
+        exit_status = _interpret_exit_status(status)
+        if status >= 0 and (status & 0x7F):
+            print(f"[Tool] Command terminated by signal {status & 0x7F}.")
+        print(f"[Tool] Exit status: {exit_status}")
+
+        return command_display, output_text, exit_status
+    finally:
+        try:
+            os.remove(temp_path)
+        except OSError:
+            pass
+
+
+def handle_tool_requests(raw_text: str, input_stream: Optional[IO[str]]) -> Tuple[str, List[str]]:
+    sanitized_text, requests = _extract_tool_requests(raw_text)
+    if not requests:
+        return raw_text.strip(), []
+
+    processed_text = sanitized_text.strip()
+    appended_chunks: List[str] = []
+    history_entries: List[str] = []
+
+    for request in requests:
+        cmd = str(request.get("cmd", "")).strip()
+        if not cmd:
+            continue
+        parms = [str(item) for item in request.get("parms", [])]
+        command_display = shlex.join([cmd, *parms])
+
+        print(f"\n[Tool Request] {command_display}")
+        is_safe, unsafe_reason = _is_command_safe(cmd, parms)
+        if not is_safe:
+            warning = "[Tool] Command blocked."
+            if unsafe_reason:
+                warning = f"[Tool] Command blocked: {unsafe_reason}"
+            print(warning)
+            note = f"[Command blocked: {command_display}]"
+            if unsafe_reason:
+                note = f"{note} {unsafe_reason}"
+            appended_chunks.append(note.strip())
+            history_reason = f"{unsafe_reason}" if unsafe_reason else "unsafe command detected"
+            history_entries.append(f"Tool: {command_display} blocked ({history_reason})")
+            continue
+
+        if not _prompt_yes_no("Execute command? [y/N]: ", input_stream):
+            print("[Tool] Command skipped.")
+            appended_chunks.append(f"[Command skipped: {command_display}]")
+            history_entries.append(f"Tool: {command_display} (skipped)")
+            continue
+
+        executed_display, output_text, exit_status = _execute_tool_command(cmd, parms)
+        summary_text, was_summarized, metrics = summarize_tool_output(output_text)
+        total_lines, total_chars = metrics
+
+        note_lines = [f"[Command executed: {executed_display} (exit {exit_status})]"]
+        history_lines = [f"Tool: {executed_display} (exit {exit_status})"]
+
+        if summary_text:
+            if was_summarized:
+                metrics_text = f"{SUMMARY_NOTICE} (original {total_lines} lines / {total_chars} chars)"
+                note_lines.append(metrics_text)
+                history_lines.append(metrics_text)
+            note_lines.append(summary_text)
+            history_lines.append(summary_text)
+        elif was_summarized:
+            metrics_text = f"{SUMMARY_NOTICE} (original {total_lines} lines / {total_chars} chars)"
+            note_lines.append(metrics_text)
+            history_lines.append(metrics_text)
+
+        appended_chunks.append("\n".join(line for line in note_lines if line).strip())
+        history_entries.append("\n".join(line for line in history_lines if line).strip())
+
+    combined_note = "\n\n".join(chunk for chunk in appended_chunks if chunk).strip()
+    if combined_note:
+        if processed_text:
+            processed_text = f"{processed_text}\n\n{combined_note}".strip()
+        else:
+            processed_text = combined_note
+
+    if not processed_text:
+        processed_text = "[Tool command processed]"
+
+    return processed_text, history_entries
+
+
 # ---------------------------------------------------------------------------
 # Interactive loop
 # ---------------------------------------------------------------------------
@@ -488,7 +947,7 @@ def interactive_mode(
     input_stream: Optional[IO[str]] = None,
 ) -> None:
     profile_info = f" (profile '{profile_name}')" if profile_name else ""
-    print(f"Interactive mode with model '{provider.model}' via {provider.base_url}{profile_info}")
+    print(f"Model '{provider.model}' via {provider.base_url}{profile_info}")
     print(f"Mode: {mode}, Temperature: {temperature} (provider: {provider.name})")
     print("Type 'exit' or Ctrl+C to quit.")
 
@@ -546,11 +1005,16 @@ def interactive_mode(
                 response_text, final_text = provider.batch(system_prompt, user_prompt, conversation_input, temperature)
                 print(response_text)
 
+            final_text, tool_history_entries = handle_tool_requests(final_text, local_stream)
+
             if memory_lines > 0:
                 if user_text:
                     memory.append(f"User: {user_text}")
                 if final_text:
                     memory.append(f"Assistant: {final_text.strip()}")
+                for entry in tool_history_entries:
+                    if entry:
+                        memory.append(entry.strip())
                 if len(memory) > memory_lines:
                     memory[:] = memory[-memory_lines:]
     finally:
@@ -565,14 +1029,64 @@ def interactive_mode(
 # Provider construction
 # ---------------------------------------------------------------------------
 
-def create_openai_client(base_url: str, require_api_key: bool, env_var: str = "OPENAI_API_KEY") -> OpenAI:
-    api_key = os.getenv(env_var)
-    if not api_key and env_var != "OPENAI_API_KEY":
-        api_key = os.getenv("OPENAI_API_KEY")
+
+def _iter_dotenv_candidates(config_source: Optional[str]) -> List[Path]:
+    discovered = find_dotenv(usecwd=True)
+    candidates: List[Path] = []
+    if discovered:
+        candidates.append(Path(discovered))
+    candidates.append(Path.cwd() / ".env")
+    if config_source:
+        source_path = Path(config_source).expanduser()
+        base_dir = source_path.parent if source_path.is_file() else source_path
+        candidates.append(base_dir / ".env")
+
+    unique: List[Path] = []
+    seen: set[str] = set()
+    for candidate in candidates:
+        try:
+            resolved = candidate.resolve()
+        except OSError:
+            continue
+        key = str(resolved)
+        if key in seen:
+            continue
+        seen.add(key)
+        if resolved.exists():
+            unique.append(resolved)
+    return unique
+
+
+def _resolve_api_key(env_var: str, config_source: Optional[str]) -> Optional[str]:
+    direct = os.getenv(env_var)
+    if direct:
+        return direct
+
+    for dotenv_path in _iter_dotenv_candidates(config_source):
+        values = dotenv_values(str(dotenv_path))
+        if not values:
+            continue
+        candidate = values.get(env_var)
+        if candidate:
+            return candidate
+        if env_var != "OPENAI_API_KEY":
+            fallback = values.get("OPENAI_API_KEY")
+            if fallback:
+                return fallback
+    return None
+
+
+def create_openai_client(
+    base_url: str,
+    require_api_key: bool,
+    env_var: str = "OPENAI_API_KEY",
+    config_source: Optional[str] = None,
+) -> OpenAI:
+    api_key = _resolve_api_key(env_var, config_source)
     if not api_key:
         if require_api_key:
             print(
-                f"[Error] {env_var} is not set. Create a .env file with {env_var}=<token> or export it.",
+                f"[Error] {env_var} is not set. Export it or add it to a .env file.",
                 file=sys.stderr,
             )
             sys.exit(1)
@@ -695,7 +1209,11 @@ def build_parser() -> argparse.ArgumentParser:
 
 def main(argv: Optional[Sequence[str]] = None) -> None:
     parser = build_parser()
-    args = parser.parse_args(argv)
+    argv_list = list(argv) if argv is not None else sys.argv[1:]
+    version = get_app_version()
+    if any(flag in argv_list for flag in ("-h", "--help")):
+        print_version_banner(version)
+    args = parser.parse_args(argv_list)
 
     config_data, config_source = load_config(args.config)
     api_config = config_data.get("api") if isinstance(config_data.get("api"), dict) else {}
@@ -733,13 +1251,20 @@ def main(argv: Optional[Sequence[str]] = None) -> None:
     else:
         require_key = provider_name == "openai"
         env_var = str(profile_settings.get("api_key_env", "OPENAI_API_KEY"))
-        client = create_openai_client(base_url, require_key, env_var)
+        client = create_openai_client(base_url, require_key, env_var, config_source)
         provider = OpenAIProvider(client, base_url, model)
 
     mode = args.mode
     if mode is None:
         mode = "batch" if not sys.stdin.isatty() else "stream"
 
+    profile_info = f"profile '{active_profile}'" if active_profile else None
+    extra_parts = [f"model '{provider.model}'", f"via {provider.base_url}"]
+    if profile_info:
+        extra_parts.append(f"({profile_info})")
+    banner_extra = " ".join(extra_parts)
+    print_version_banner(version, banner_extra)
+
     if not sys.stdin.isatty():
         data = sys.stdin.read()
         if not data.strip():
@@ -754,21 +1279,40 @@ def main(argv: Optional[Sequence[str]] = None) -> None:
             )
             return
 
+        summarized_input, input_was_summarized, input_metrics = summarize_user_input(data)
+        if not summarized_input:
+            summarized_input = data.strip()
+
+        conversation_input = summarized_input
+        input_summary_notice = ""
+        if input_was_summarized:
+            lines, chars = input_metrics
+            input_summary_notice = f"{USER_SUMMARY_NOTICE} (original {lines} lines / {chars} chars)"
+            print(input_summary_notice)
+            if summarized_input:
+                conversation_input = f"{input_summary_notice}\n{summarized_input}"
+            else:
+                conversation_input = input_summary_notice
+
         final_text = ""
         if mode == "stream":
             print("LLM (thinking): ", end="", flush=True)
-            final_text = provider.stream(system_prompt, user_prompt, data, temperature)
+            final_text = provider.stream(system_prompt, user_prompt, conversation_input, temperature)
         else:
-            response_text, final_text = provider.batch(system_prompt, user_prompt, data, temperature)
+            response_text, final_text = provider.batch(system_prompt, user_prompt, conversation_input, temperature)
             print(response_text)
 
+        final_text, tool_history_entries = handle_tool_requests(final_text, None)
+
         if chat_after_stdin:
             memory_seed: List[str] = []
-            data_text = data.strip()
-            if data_text:
-                memory_seed.append(f"User: {data_text}")
+            if conversation_input:
+                memory_seed.append(f"User: {conversation_input.strip()}")
             if final_text:
                 memory_seed.append(f"Assistant: {final_text.strip()}")
+            for entry in tool_history_entries:
+                if entry:
+                    memory_seed.append(entry.strip())
             if memory_lines > 0 and memory_seed:
                 memory_seed = memory_seed[-memory_lines:]
             interactive_mode(
